/*
 * vcftools.h
 */

#ifndef VCFTOOLS_H_
#define VCFTOOLS_H_

#include "output_log.h"
#include "parameters.h"
#include "bcf_file.h"
#include "vcf_file.h"
#include "variant_file.h"
#include "header.h"

#endif /* VCFTOOLS_H_ */
